//
// Left Leaning Red Black tree with relative node keys, i.e. the key in each node
// is relative to its immediate neighbors.
//
// When using this code please cite:
//
// M. Holenderski, R.J. Bril and J.J. Lukkien, "Red-black trees with relative node keys",
// Information Processing Letters, vol. 114, issue 11, November 2014.
//
// Copyright 2013 Mike Holenderski, mike@holenderski.com
//

#include "llrb.h"

#define RED 0
#define BLK 1

#define kRootTime 1000

typedef struct llrb_node_s {
  int time;
  struct llrb_node_s* left;
  struct llrb_node_s* right;
  int color;
  int info;
} Node;

typedef enum {
  kNodeKindRoot = 0,
  kNodeKindLeft,
  kNodeKindRight
} NodeKind;

void init(LLRBTree *tree, LLRB_COMPARISON comparison) {
  tree->root = NULL;
  tree->comparison = comparison;
}

static int do_max_depth(Node *h) {
  int l, r;
  
  if (h == NULL)
    return 0;
  l = do_max_depth(h->left);
  r = do_max_depth(h->right);
  return (l > r) ? l + 1 : r + 1;
}

int max_depth(LLRBTree *tree) {
  return do_max_depth(tree->root);
}

void *safe_malloc(size_t n) {
  void *obj = malloc(n);
  if (obj == NULL) {
    fprintf(stderr, "out of memory\n");
    exit(1);
  }
  return obj;
}

static Node *new_node(int time) {
  Node *node = safe_malloc(sizeof *node);
  node->time = time;
  node->left = node->right = NULL;
  node->color = RED;
  return node;
}

static int isRed(Node *h) {
  return h != NULL && h->color == RED;
}

static Node *rotateLeft(LLRBTree *tree, Node *h, NodeKind kind, int* cumulative) {
//  printf("rotate left %d kind %d cumulative %d\n", h->info, kind, *cumulative);
  /* a = h, b = h->right, x = h->right->left */

  if (h->right) {
    int a = h->time;
    int b = h->right->time;
    /* common: r'(a) = r(b), r'(x) = r(b) - r(x); */
    h->time = b;
    if (h->right->left) { h->right->left->time = b - h->right->left->time; }

    switch (kind) {
      case kNodeKindRight:
        /* r'(b) = r(a) + r(b) */
        h->right->time += a;
        break;
        
      case kNodeKindLeft:
        /* r'(b) = r(a) - r(b) */
        h->right->time = a - b;
        break;
  
      case kNodeKindRoot:
        /* c'(root) = c(root) + r(b) */
        tree->cumulative += b;
        tree->root = h->right;
        // note that the root relative time is never used, but we set it for debuging
        h->right->time = kRootTime; 
        break;
    }
    *cumulative += b;
  }
  
  Node *x = h->right;
  h->right = x->left;
  x->left = h;
  x->color = h->color;
  h->color = RED;
  return x;
}

static Node *rotateRight(LLRBTree *tree, Node *h, NodeKind kind, int* cumulative) {
//  printf("rotate right %d kind %d cumulative %d\n", h->info, kind, *cumulative);
  /* a = h->left, b = h, x = h->left->right */

  if (h->left) {
    int a = h->left->time;
    int b = h->time;
    /* common: r'(b) = r(a), r'(x) = r(a) - r(x) */
    h->time = a;
    if (h->left->right) { h->left->right->time = a - h->left->right->time; }

    switch (kind) {
      case kNodeKindRight:
        /* r'(a) := r(b) - r(a) */
        h->left->time = b - a;
        break;
        
      case kNodeKindLeft:
        /* r'(a) := r(a) + r(b) */
        h->left->time += b;
        break;
  
      case kNodeKindRoot:
        /* c'(root) := c(root) - r(a) */
        tree->cumulative -= a;
        tree->root = h->left;
        // note that the root relative time is never used, but we set it for debuging
        h->left->time = kRootTime;
        break;
    }
    *cumulative -= a;
  }
  
  Node *x = h->left;
  h->left = x->right;
  x->right = h;
  x->color = h->color;
  h->color = RED;
  return x;
}

static void flip(Node *h) {
  h->color ^= 1;
  h->left->color ^= 1;
  h->right->color ^= 1;
}

static Node *do_insert(LLRBTree *tree, Node *h, int time, int cumulative, NodeKind kind, int info) {
  if (h == NULL) {
    Node* node = new_node(time);
    node->info = info;
    return node;
  }
  
  if (isRed(h->left) && isRed(h->right)) { flip(h); }
  
  int cmp = tree->comparison(time, cumulative);
  if (cmp < 0) {
    if (!h->left) {
      h->left = do_insert(tree, h->left, cumulative - time, 0, kNodeKindLeft, info);
    } else {
      h->left = do_insert(tree, h->left, time, cumulative - h->left->time, kNodeKindLeft, info);    
    }
  } else if (cmp >= 0) {
    if (!h->right) {
      h->right = do_insert(tree, h->right, time - cumulative, 0, kNodeKindRight, info);
    } else {
      h->right = do_insert(tree, h->right, time, cumulative + h->right->time, kNodeKindRight, info);
    }
  } else {
    h->time = time;
  }
  
  if (isRed(h->right) && !isRed(h->left)) {
    h = rotateLeft(tree, h, kind, &cumulative);
  }
  if (isRed(h->left) && isRed(h->left->left)) {
    h = rotateRight(tree, h, kind, &cumulative);
  }  
  return h;
}

void insert(LLRBTree *tree, int time) {
  printf("insert %d\n", time);

  tree->root = do_insert(tree, tree->root, time, tree->cumulative, kNodeKindRoot, time);
  tree->root->color = BLK;
  if (!tree->root->left && !tree->root->right) {
    tree->cumulative = time;
    tree->root->time = kRootTime;
  }
  printTree(tree->file, tree);
}

void printNode(FILE* f, Node* node) {
  fprintf(f, "{%d ", node->time);
  if (node->left) printNode(f, node->left); else fprintf(f, "{}");
  fprintf(f, " ");
  if (node->right) printNode(f, node->right); else fprintf(f, "{}");
  fprintf(f, " %s %d} ", node->color ? "black" : "red", node->info);
}

void printTree(FILE* f, LLRBTree *tree) {
  printNode(f, tree->root);
  fprintf(f, " {c(root) = %d}\n", tree->cumulative);
}

Node* moveRedLeft(LLRBTree* tree, Node* h, NodeKind kind, int* cumulative)  {
  printf("move left %d kind %d cumulative %d\n", h->info, kind, *cumulative);
  flip(h);
  if (isRed(h->right->left)) {   
    int rightCumulative = *cumulative + h->right->time;
    h->right = rotateRight(tree, h->right, kNodeKindRight, &rightCumulative);
    h = rotateLeft(tree, h, kind, cumulative);
    flip(h);
  }
  return h;
}

Node* moveRedRight(LLRBTree* tree, Node* h, NodeKind kind, int* cumulative) {
  printf("move right %d kind %d cumulative %d\n", h->info, kind, *cumulative);
  flip(h);
  if (isRed(h->left->left)) {
    h = rotateRight(tree, h, kind, cumulative);
    flip(h);
  }
  return h;
}

Node* fixUp(LLRBTree* tree, Node* h, NodeKind kind, int* cumulative) {
  if (isRed(h->right)) { h = rotateLeft(tree, h, kind, cumulative); }
  if (h->left && isRed(h->left) && isRed(h->left->left)) { h = rotateRight(tree, h, kind, cumulative); }
  if (isRed(h->left) && isRed(h->right)) { flip(h); }
  return h;
}

Node* deleteMin(LLRBTree* tree, Node* h, NodeKind kind, int* cumulative) {
//  printf("deleteMin %d kind %d cumulative %d\n", h->info, kind, *cumulative);
  if (!h->left) return NULL;
  if (!isRed(h->left) && !isRed(h->left->left)) {
    h = moveRedLeft(tree, h, kind, cumulative);
  }
  int leftCumulative = *cumulative - h->left->time;
  h->left = deleteMin(tree, h->left, kNodeKindLeft, &leftCumulative); 
  return fixUp(tree, h, kind, cumulative);
}

Node* do_delete(LLRBTree* tree, Node* h, int time, int cumulative, NodeKind kind) {
  if (!h) { return NULL; }

  if (tree->comparison(time, cumulative) < 0) {
    if (!h->left) { return h; }
    if (!isRed(h->left) && !isRed(h->left->left)) {
      h = moveRedLeft(tree, h, kind, &cumulative);
    }
    h->left = do_delete(tree, h->left, time, cumulative - h->left->time, kNodeKindLeft);
  } else {
    if (isRed(h->left)) {
      h = rotateRight(tree, h, kind, &cumulative);
    }
    if (tree->comparison(time, cumulative) == 0 && !h->right) { return NULL; }
    if (!isRed(h->right) && !isRed(h->right->left)) {
      h = moveRedRight(tree, h, kind, &cumulative);
    }
    if (tree->comparison(time, cumulative) == 0) {
      // find the minimum in the right subtree and its relative time to h
      int Rm = h->right->time;
      Node* min;
      for (min = h->right; min->left; min = min->left) {
        if (min->left) { Rm -= min->left->time; }
      }
      h->info = min->info;
      // adjust the relative times of self and children
      switch (kind) {
        case kNodeKindLeft:
          /* r'(a) = r(a) - R(m) */
          h->time -= Rm;
          break;
        case kNodeKindRight:
          /* r'(a) = r(a) + R(m) */
          h->time += Rm;
          break;
        case kNodeKindRoot:
          /* c'(root) = c(root) + R(m) */
          tree->cumulative += Rm;
          h->time += kRootTime;
          break;
      }
      cumulative += Rm;
      if (h->left) { h->left->time += Rm; }
      if (h->right) { h->right->time -= Rm; }
      int rightCumulative = cumulative + h->right->time;
      h->right = deleteMin(tree, h->right, kNodeKindRight, &rightCumulative);
    } else {
      h->right = do_delete(tree, h->right, time, cumulative + h->right->time, kNodeKindRight);
    }
  }

  return fixUp(tree, h, kind, &cumulative);
}

void delete(LLRBTree* tree, int time) {
  printf("delete %d\n", time);
  tree->root = do_delete(tree, tree->root, time, tree->cumulative, kNodeKindRoot);
  if (tree->root) {
    tree->root->color = BLK;
  }
  printTree(tree->file, tree);
}
